import logging

from fsspec import AbstractFileSystem
from deltacat import logs
import deltacat.compute.converter.utils.iceberg_columns as sc
import daft
from deltacat.utils.daft import _get_s3_io_config
from daft import TimeUnit, DataFrame
import pyarrow as pa
from typing import Callable, Optional, List, Dict, Any
from deltacat.utils.pyarrow import sliced_string_cast
from deltacat.compute.converter.constants import IDENTIFIER_FIELD_DELIMITER
from deltacat.compute.converter.utils.s3u import upload_table_with_retry
from pyiceberg.manifest import DataFile
import pyarrow.compute as pc
from deltacat.types.media import ContentType
from deltacat.types.tables import (
    get_table_writer,
    get_table_slicer,
    write_sliced_table as types_write_sliced_table,
)
from deltacat.storage import LocalTable, DistributedDataset
from typing import Union

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def download_data_table_and_append_iceberg_columns(
    file: DataFile,
    columns_to_download: List[str],
    additional_columns_to_append: Optional[List[str]] = [],
    s3_client_kwargs: Optional[Dict[str, Any]] = None,
) -> pa.Table:
    table = download_parquet_with_daft_hash_applied(
        identifier_columns=columns_to_download,
        file=file,
        s3_client_kwargs=s3_client_kwargs,
    )

    if sc._FILE_PATH_COLUMN_NAME in additional_columns_to_append:
        table = sc.append_file_path_column(table, file.file_path)
    if sc._ORDERED_RECORD_IDX_COLUMN_NAME in additional_columns_to_append:
        record_idx_iterator = iter(range(len(table)))
        table = sc.append_record_idx_col(table, record_idx_iterator)

    return table


def download_parquet_with_daft_hash_applied(
    identifier_columns: List[str],
    file: DataFile,
    s3_client_kwargs: Optional[Dict[str, Any]],
    **kwargs: Any,
) -> pa.Table:

    # TODO: Add correct read kwargs as in:
    #  https://github.com/ray-project/deltacat/blob/383855a4044e4dfe03cf36d7738359d512a517b4/deltacat/utils/daft.py#L97

    coerce_int96_timestamp_unit = TimeUnit.from_str(
        kwargs.get("coerce_int96_timestamp_unit", "ms")
    )

    # TODO: Use Daft SHA1 hash instead to minimize probably of data corruption
    io_config = _get_s3_io_config(s3_client_kwargs=s3_client_kwargs)
    df = daft_read_parquet(
        path=file.file_path,
        io_config=io_config,
        coerce_int96_timestamp_unit=coerce_int96_timestamp_unit,
    )

    hash_column = concatenate_hashed_identifier_columns(
        df=df, identifier_columns=identifier_columns
    )

    table = pa.Table.from_arrays(
        [hash_column], names=[sc._IDENTIFIER_COLUMNS_HASH_COLUMN_NAME]
    )

    return table


def daft_read_parquet(
    path: str, io_config: Dict[str, Any], coerce_int96_timestamp_unit: TimeUnit
) -> DataFrame:
    df = daft.read_parquet(
        path=path,
        io_config=io_config,
        coerce_int96_timestamp_unit=coerce_int96_timestamp_unit,
    )
    return df


def concatenate_hashed_identifier_columns(
    df: DataFrame, identifier_columns: List[str]
) -> pa.Array:
    pk_hash_columns = []
    previous_hash_column_length = None
    for i in range(len(identifier_columns)):
        pk_hash_column = df.select(daft.col(identifier_columns[i]).hash())
        pk_hash_column_arrow = pk_hash_column.to_arrow()

        # Assert that each hash column downloaded are same length to ensure we don't create mismatch between columns.
        if not previous_hash_column_length:
            previous_hash_column_length = len(pk_hash_column_arrow)
        else:
            assert previous_hash_column_length == len(pk_hash_column_arrow), (
                f"Identifier column Length mismatch: {identifier_columns[i]} has length {len(pk_hash_column_arrow)} "
                f"but expected {previous_hash_column_length}."
            )
            previous_hash_column_length = len(pk_hash_column_arrow)

        # Convert identifier from different datatypes to string here
        pk_hash_column_str = sliced_string_cast(
            pk_hash_column_arrow[identifier_columns[i]]
        )
        assert len(pk_hash_column_str) == previous_hash_column_length, (
            f"Casting column Length mismatch: {identifier_columns[i]} has length {len(pk_hash_column_str)} after casting, "
            f"before casting length: {previous_hash_column_length}."
        )

        pk_hash_columns.append(pk_hash_column_str)

    pk_hash_columns.append(IDENTIFIER_FIELD_DELIMITER)
    pk_hash_columns_concatenated = pc.binary_join_element_wise(
        *pk_hash_columns, null_handling="replace"
    )
    assert len(pk_hash_columns_concatenated) == previous_hash_column_length, (
        f"Concatenated column Length mismatch: Final concatenated identifier column has length {len(pk_hash_columns_concatenated)}, "
        f"before concatenating length: {previous_hash_column_length}."
    )

    return pk_hash_columns_concatenated


def write_sliced_table(
    table: Union[LocalTable, DistributedDataset],
    base_path: str,
    table_writer_kwargs: Optional[Dict[str, Any]],
    content_type: ContentType = ContentType.PARQUET,
    max_records_per_file: Optional[int] = 4000000,
    filesystem: Optional[Union[AbstractFileSystem, pa.fs.FileSystem]] = None,
    **kwargs,
) -> List[str]:
    """
    Writes the given table to 1 or more files and return the paths
    of the files written.
    """
    if isinstance(filesystem, pa.fs.FileSystem):
        table_writer_fn = get_table_writer(table)
        table_slicer_fn = get_table_slicer(table)

        # Create a wrapper for the table writer that ensures directory creation
        def table_writer_with_dir_creation(
            dataframe: Any,
            base_path: str,
            filesystem: Optional[Union[AbstractFileSystem, pa.fs.FileSystem]],
            block_path_provider: Callable,
            content_type: str = ContentType.PARQUET.value,
            **kwargs,
        ):
            try:
                # Ensure base path directory exists
                if isinstance(base_path, str):
                    # Normalize the base path and ensure it's treated as a directory path
                    base_dir = base_path.rstrip("/")
                    filesystem.create_dir(base_dir, recursive=True)
            except Exception:
                # Directory might already exist or there might be permission issues
                # Let the original write attempt proceed
                pass
            return table_writer_fn(
                dataframe,
                base_path,
                filesystem,
                block_path_provider,
                content_type,
                **kwargs,
            )

        # TODO(pdames): Disable redundant file info fetch currently
        #   used to construct unused manifest entry metadata.
        manifest_entry_list = types_write_sliced_table(
            table=table,
            base_path=base_path,
            filesystem=filesystem,
            max_records_per_entry=max_records_per_file,
            table_writer_fn=table_writer_with_dir_creation,
            table_slicer_fn=table_slicer_fn,
            table_writer_kwargs=table_writer_kwargs,
            content_type=content_type,
        )
        paths = [entry.uri for entry in manifest_entry_list]
        return paths
    else:
        return upload_table_with_retry(
            table=table,
            s3_url_prefix=base_path,
            s3_table_writer_kwargs=table_writer_kwargs,
            content_type=content_type,
            max_records_per_file=max_records_per_file,
            s3_file_system=filesystem,
            **kwargs,
        )



import ray
import pytest
import tempfile
import shutil
import pandas as pd
from deltacat.storage import metastore
from deltacat.catalog import CatalogProperties
from deltacat.types.media import ContentType
from deltacat.storage.model.types import DeltaType
from deltacat.compute.compactor_v2.compaction_session import compact_partition
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from deltacat.compute.resource_estimation import ResourceEstimationMethod
from deltacat.exceptions import ValidationError
from deltacat.tests.compute.test_util_common import (
    get_rci_from_partition,
    read_audit_file,
)


@pytest.fixture(autouse=True, scope="module")
def setup_ray_cluster():
    ray.init(local_mode=True, ignore_reinit_error=True)
    yield
    ray.shutdown()


@pytest.fixture
def catalog():
    """Create a temporary catalog for testing."""
    tmpdir = tempfile.mkdtemp()
    catalog = CatalogProperties(root=tmpdir)
    yield catalog
    shutil.rmtree(tmpdir)


class TestCompactionSessionMain:
    """Compaction session tests using main deltacat metastore."""

    NAMESPACE = "compact_partition_main_test"
    ERROR_RATE = 0.05

    # Test data equivalent to the CSV files
    BACKFILL_DATA = pd.DataFrame(
        {
            "pk": ["2022-10-21", "2022-10-20", "2022-11-24", "2023-10-23"],
            "value": [1, 2, 3, 4],
        }
    )

    INCREMENTAL_DATA = pd.DataFrame(
        {"pk": ["2022-10-21", "2022-11-25"], "value": [1, 5]}
    )

    def _create_namespace_and_table(self, namespace_suffix, catalog):
        """Helper to create namespace and table for tests."""
        namespace_name = f"{self.NAMESPACE}_{namespace_suffix}"

        # Create namespace
        namespace = metastore.create_namespace(
            namespace=namespace_name,
            catalog=catalog,
        )

        # Create table and table version
        table, table_version, stream = metastore.create_table_version(
            namespace=namespace.locator.namespace,
            table_name=f"table_{namespace_suffix}",
            catalog=catalog,
        )

        return namespace, table, table_version, stream

    def _stage_and_commit_partition(self, stream, catalog):
        """Helper to stage and commit a partition."""
        partition = metastore.stage_partition(
            stream=stream,
            catalog=catalog,
        )
        return metastore.commit_partition(
            partition=partition,
            catalog=catalog,
        )

    def _stage_and_commit_delta(
        self, data, partition, catalog, delta_type=DeltaType.UPSERT
    ):
        """Helper to stage and commit a delta with data."""
        staged_delta = metastore.stage_delta(
            data=data,
            partition=partition,
            catalog=catalog,
            content_type=ContentType.PARQUET,
            delta_type=delta_type,
        )

        return metastore.commit_delta(
            delta=staged_delta,
            catalog=catalog,
        )

    def test_compact_partition_basic_sanity(self, catalog):
        """Basic sanity test to verify compact_partition works with main metastore."""

        # Create source namespace and table
        source_namespace = metastore.create_namespace(
            namespace=f"{self.NAMESPACE}_source",
            catalog=catalog,
        )

        # Create destination namespace and table
        dest_namespace = metastore.create_namespace(
            namespace=f"{self.NAMESPACE}_dest",
            catalog=catalog,
        )

        # Create a simple test dataset
        test_data = pd.DataFrame(
            {
                "pk": [1, 2, 3, 4],
                "name": ["A", "B", "C", "D"],
                "value": [10, 20, 30, 40],
            }
        )

        # Create source table and partition
        (
            source_table,
            source_table_version,
            source_stream,
        ) = metastore.create_table_version(
            namespace=source_namespace.locator.namespace,
            table_name="source_table",
            catalog=catalog,
        )

        source_partition = metastore.stage_partition(
            stream=source_stream,
            catalog=catalog,
        )
        source_partition = metastore.commit_partition(
            partition=source_partition,
            catalog=catalog,
        )

        # Stage and commit a delta to the source partition
        staged_delta = metastore.stage_delta(
            data=test_data,
            partition=source_partition,
            catalog=catalog,
            content_type=ContentType.PARQUET,
            delta_type=DeltaType.UPSERT,
        )

        source_delta = metastore.commit_delta(
            delta=staged_delta,
            catalog=catalog,
        )

        # Create destination table and partition
        dest_table, dest_table_version, dest_stream = metastore.create_table_version(
            namespace=dest_namespace.locator.namespace,
            table_name="dest_table",
            catalog=catalog,
        )

        dest_partition = metastore.stage_partition(
            stream=dest_stream,
            catalog=catalog,
        )
        dest_partition = metastore.commit_partition(
            partition=dest_partition,
            catalog=catalog,
        )
        # Test compact_partition with minimal parameters
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 1,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["pk"],
                    "all_column_names": ["pk", "name", "value"],
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": source_partition.locator,
                }
            )
        )

        # Basic verification - if we get here without exceptions, the basic flow works

        # Get a fresh reference to the destination partition to see updates
        updated_dest_partition = metastore.get_partition(
            stream_locator=dest_stream.locator,
            partition_values=None,  # unpartitioned
            catalog=catalog,
        )

        print(
            f"Original destination partition stream position: {dest_partition.stream_position}"
        )
        print(
            f"Updated destination partition stream position: {updated_dest_partition.stream_position}"
        )

        # Verify that the destination partition now has some deltas
        dest_partition_deltas = metastore.list_partition_deltas(
            partition_like=updated_dest_partition,
            include_manifest=True,
            catalog=catalog,
        )

        delta_count = len(dest_partition_deltas.all_items())
        print(f"Found {delta_count} delta(s) in destination partition")

        # Verify that at least one compacted delta was written to the destination partition
        assert (
            delta_count > 0
        ), f"Expected at least one delta in destination partition, but found {delta_count}"

        # Print some info about the delta(s) found
        for i, delta in enumerate(dest_partition_deltas.all_items()):
            print(
                f"Delta {i+1}: stream_position={delta.stream_position}, type={delta.type}, record_count={delta.meta.record_count if delta.meta else 'N/A'}"
            )

        print(
            f"✅ Basic sanity test PASSED! compact_partition works with main deltacat metastore and wrote {delta_count} delta(s) to destination partition."
        )

    def test_compact_partition_when_no_input_deltas_to_compact(self, catalog):
        """Test compaction when there are no input deltas to compact."""
        # Create source and destination namespaces/tables
        _, _, _, source_stream = self._create_namespace_and_table("source", catalog)
        _, _, _, dest_stream = self._create_namespace_and_table("destination", catalog)

        # Create source and destination partitions (no deltas)
        source_partition = self._stage_and_commit_partition(source_stream, catalog)
        dest_partition = self._stage_and_commit_partition(dest_stream, catalog)

        # For partitions with no deltas, use stream position 0 or 1 as the last position to compact
        last_position = source_partition.stream_position or 0

        # Attempt compaction
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": last_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["pk"],
                    "all_column_names": ["pk", "value"],
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": source_partition.locator,
                }
            )
        )

    def test_compact_partition_when_incremental_then_rci_stats_accurate(self, catalog):
        """Test case which asserts the RCI stats are correctly generated for a rebase and incremental use-case."""
        # Create source and destination namespaces/tables
        _, _, _, source_stream = self._create_namespace_and_table("source", catalog)
        _, _, _, dest_stream = self._create_namespace_and_table("destination", catalog)

        # Create source partition and commit backfill data
        source_partition = self._stage_and_commit_partition(source_stream, catalog)
        source_delta = self._stage_and_commit_delta(
            self.BACKFILL_DATA, source_partition, catalog
        )

        # Create destination partition
        dest_partition = self._stage_and_commit_partition(dest_stream, catalog)

        # First compaction with backfill data
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["pk"],
                    "all_column_names": ["pk", "value"],
                    "original_fields": {"pk", "value"},
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": source_delta.partition_locator,
                }
            )
        )

        # Get RoundCompletionInfo from the compacted partition instead of file
        backfill_rci = get_rci_from_partition(
            dest_partition.locator, metastore, catalog=catalog
        )
        # Get catalog root for audit file resolution
        catalog_root = catalog.root

        compaction_audit = CompactionSessionAuditInfo(
            **read_audit_file(backfill_rci.compaction_audit_url, catalog_root)
        )

        # Verify that inflation and record size values are reasonable (not exact due to storage differences)
        # Note: inflation values may be None in some storage implementations
        if backfill_rci.input_inflation is not None:
            assert (
                0.01 <= backfill_rci.input_inflation <= 0.2
            )  # Reasonable inflation range
        if backfill_rci.input_average_record_size_bytes is not None:
            assert (
                5 <= backfill_rci.input_average_record_size_bytes <= 50
            )  # Reasonable record size range

        assert compaction_audit.input_records == 4
        assert compaction_audit.records_deduped == 0
        assert compaction_audit.records_deleted == 0
        assert compaction_audit.untouched_file_count == 0
        assert compaction_audit.untouched_record_count == 0
        assert compaction_audit.untouched_size_bytes == 0
        assert compaction_audit.untouched_file_ratio == 0
        assert compaction_audit.uniform_deltas_created == 1
        assert compaction_audit.hash_bucket_count == 2
        assert compaction_audit.input_file_count == 1
        assert compaction_audit.output_file_count == 2
        # Allow larger tolerance for file size differences between storage implementations
        # File sizes can vary significantly due to different compression, metadata, etc.
        assert compaction_audit.output_size_bytes > 0
        assert compaction_audit.input_size_bytes > 0

        # Now commit incremental data and run incremental compaction
        new_source_delta = self._stage_and_commit_delta(
            self.INCREMENTAL_DATA, source_partition, catalog
        )

        # Use the original destination partition for incremental compaction
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": new_source_delta.stream_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["pk"],
                    "all_column_names": ["pk", "value"],
                    "original_fields": {"pk", "value"},
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": new_source_delta.partition_locator,
                }
            )
        )

        # Get RoundCompletionInfo from the compacted partition instead of file
        new_rci = get_rci_from_partition(
            dest_partition.locator, metastore, catalog=catalog
        )
        # Get catalog root for audit file resolution
        catalog_root = catalog.root

        compaction_audit = CompactionSessionAuditInfo(
            **read_audit_file(new_rci.compaction_audit_url, catalog_root)
        )

        # Verify incremental compaction metrics are reasonable (looser bounds due to storage differences)
        # Note: inflation values may be None in some storage implementations
        if new_rci.input_inflation is not None:
            assert 0.01 <= new_rci.input_inflation <= 0.2  # Reasonable inflation range
        if new_rci.input_average_record_size_bytes is not None:
            assert (
                5 <= new_rci.input_average_record_size_bytes <= 50
            )  # Reasonable record size range

        assert compaction_audit.input_records >= 4  # At least the backfill records
        assert compaction_audit.records_deduped >= 0
        assert compaction_audit.records_deleted == 0
        assert compaction_audit.untouched_file_count >= 0
        assert compaction_audit.untouched_record_count >= 0
        # Allow larger tolerance for size differences
        assert compaction_audit.untouched_file_ratio >= 0
        assert compaction_audit.uniform_deltas_created >= 1
        assert compaction_audit.hash_bucket_count == 2
        assert compaction_audit.input_file_count >= 1
        assert compaction_audit.output_file_count >= 1
        # Allow larger tolerance for file size differences between storage implementations
        # File sizes can vary significantly due to different compression, metadata, etc.
        assert compaction_audit.output_size_bytes > 0
        assert compaction_audit.input_size_bytes > 0

    def test_compact_partition_when_hash_bucket_count_changes_then_validation_error(
        self, catalog
    ):
        """Test that changing hash bucket count between compactions raises ValidationError."""
        # Create source and destination namespaces/tables
        _, _, _, source_stream = self._create_namespace_and_table("source", catalog)
        _, _, _, dest_stream = self._create_namespace_and_table("destination", catalog)

        # Create source partition and commit backfill data
        source_partition = self._stage_and_commit_partition(source_stream, catalog)
        source_delta = self._stage_and_commit_delta(
            self.BACKFILL_DATA, source_partition, catalog
        )

        # Create destination partition
        dest_partition = self._stage_and_commit_partition(dest_stream, catalog)

        # First compaction with hash_bucket_count=2
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["pk"],
                    "all_column_names": ["pk", "value"],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": source_delta.partition_locator,
                }
            )
        )

        # Now commit incremental data and run incremental compaction with different hash bucket count
        new_source_delta = self._stage_and_commit_delta(
            self.INCREMENTAL_DATA, source_partition, catalog
        )

        # This should raise ValidationError due to hash bucket count mismatch (2 vs 1)
        with pytest.raises(ValidationError) as exc_info:
            compact_partition(
                CompactPartitionParams.of(
                    {
                        "catalog": catalog,
                        "compacted_file_content_type": ContentType.PARQUET,
                        "dd_max_parallelism_ratio": 1.0,
                        "deltacat_storage": metastore,
                        "deltacat_storage_kwargs": {"catalog": catalog},
                        "destination_partition_locator": dest_partition.locator,
                        "drop_duplicates": True,
                        "hash_bucket_count": 1,  # Different from initial compaction (2)
                        "last_stream_position_to_compact": new_source_delta.stream_position,
                        "list_deltas_kwargs": {
                            "catalog": catalog,
                            "equivalent_table_types": [],
                        },
                        "primary_keys": ["pk"],
                        "all_column_names": ["pk", "value"],
                        "rebase_source_partition_locator": None,
                        "rebase_source_partition_high_watermark": None,
                        "records_per_compacted_file": 4000,
                        "source_partition_locator": new_source_delta.partition_locator,
                    }
                )
            )

        # Verify the error message contains the expected hash bucket count mismatch details
        error_message = str(exc_info.value)
        assert "Partition hash bucket count for compaction has changed" in error_message
        assert "Hash bucket count in RCI=2" in error_message
        assert "hash bucket count in params=1" in error_message

    def test_compact_partition_when_incremental_then_intelligent_estimation_sanity(
        self, catalog
    ):
        """Test case which asserts the RCI stats are correctly generated for a rebase and incremental use-case with intelligent estimation."""
        # Create source and destination namespaces/tables
        _, _, _, source_stream = self._create_namespace_and_table("source", catalog)
        _, _, _, dest_stream = self._create_namespace_and_table("destination", catalog)

        # Create source partition and commit backfill data
        source_partition = self._stage_and_commit_partition(source_stream, catalog)
        source_delta = self._stage_and_commit_delta(
            self.BACKFILL_DATA, source_partition, catalog
        )

        # Create destination partition
        dest_partition = self._stage_and_commit_partition(dest_stream, catalog)

        # Test compaction with intelligent estimation
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["pk"],
                    "all_column_names": ["pk", "value"],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": source_delta.partition_locator,
                    "resource_estimation_method": ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
                }
            )
        )

    def test_compact_partition_when_incremental_then_content_type_meta_estimation_sanity(
        self, catalog
    ):
        """Test case which asserts the RCI stats are correctly generated for a rebase and incremental use-case with content type meta estimation."""
        # Create source and destination namespaces/tables
        _, _, _, source_stream = self._create_namespace_and_table("source", catalog)
        _, _, _, dest_stream = self._create_namespace_and_table("destination", catalog)

        # Create source partition and commit backfill data
        source_partition = self._stage_and_commit_partition(source_stream, catalog)
        source_delta = self._stage_and_commit_delta(
            self.BACKFILL_DATA, source_partition, catalog
        )

        # Create destination partition
        dest_partition = self._stage_and_commit_partition(dest_stream, catalog)

        # Test compaction with content type meta estimation
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["pk"],
                    "all_column_names": ["pk", "value"],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": source_delta.partition_locator,
                    "resource_estimation_method": ResourceEstimationMethod.CONTENT_TYPE_META,
                }
            )
        )

    def test_compact_partition_when_incremental_then_previous_inflation_estimation_sanity(
        self, catalog
    ):
        """Test case which asserts the RCI stats are correctly generated for a rebase and incremental use-case with previous inflation estimation."""
        # Create source and destination namespaces/tables
        _, _, _, source_stream = self._create_namespace_and_table("source", catalog)
        _, _, _, dest_stream = self._create_namespace_and_table("destination", catalog)

        # Create source partition and commit backfill data
        source_partition = self._stage_and_commit_partition(source_stream, catalog)
        source_delta = self._stage_and_commit_delta(
            self.BACKFILL_DATA, source_partition, catalog
        )

        # Create destination partition
        dest_partition = self._stage_and_commit_partition(dest_stream, catalog)

        # Test compaction with previous inflation estimation
        compact_partition(
            CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": {"catalog": catalog},
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        "catalog": catalog,
                        "equivalent_table_types": [],
                    },
                    "primary_keys": ["pk"],
                    "all_column_names": ["pk", "value"],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "source_partition_locator": source_delta.partition_locator,
                    "resource_estimation_method": ResourceEstimationMethod.PREVIOUS_INFLATION,
                }
            )
        )
